package jp.sourceforge.acerola3d.a3;

import java.awt.event.*;
import javax.vecmath.*;

/**
 * SIMPLEモードのコントローラ。
 */
class SimpleController extends A3Controller {
    //注視点
    Tuple3d fixationPoint= null;
    double distance;
    Quat4d cameraQuat = null;
    int lastMouseX;
    int lastMouseY;
    /**
     * 初期化します。
     */
    public void init() {
        distance = a3canvas.getCameraScale()*3.0;
        cameraQuat = a3canvas.getCameraQuat();
        Quat4d cameraQuatC = new Quat4d();
        cameraQuatC.conjugate(cameraQuat);
        Quat4d tmpQ = new Quat4d(cameraQuat);
        tmpQ.mul(new Quat4d(0.0,0.0,1.0,0.0));
        tmpQ.mul(cameraQuatC);
        tmpQ.scale(distance);
        Tuple3d tmpV = a3canvas.getCameraLoc();
        tmpV.sub(new Vector3d(tmpQ.x,tmpQ.y,tmpQ.z));
        fixationPoint = tmpV;
System.out.println(fixationPoint.toString());
System.out.println(a3canvas.getCameraLoc());
        setCamera();
    }
    /**
     * マウスが押された時のイベントをキャッチします。
     */
    public void mousePressed(A3Event ae) {
        MouseEvent me = ae.getMouseEvent();
        lastMouseX = me.getX();
        lastMouseY = me.getY();
    }

    /**
     * マウスがドラッグされた時のイベントをキャッチします。
     */
    public void mouseDragged(A3Event ae) {
        MouseEvent me = ae.getMouseEvent();
        int shiftX = me.getX() - lastMouseX;
        int shiftY = me.getY() - lastMouseY;
        if ((me.getModifiersEx()&MouseEvent.BUTTON1_DOWN_MASK)!=0) {
            //回転
            Quat4d q0 = new Quat4d(0.0,0.0,0.0,1.0);
            double t = - shiftX * 0.01;
            Quat4d q1 = new Quat4d(0.0,Math.sin(t),0.0,Math.cos(t));
            t = - shiftY * 0.01;
            Quat4d q2 = new Quat4d(Math.sin(t),0.0,0.0,Math.cos(t));
            q0.mul(q1);
            q0.mul(q2);
            cameraQuat.mul(q0);
        } else if ((me.getModifiersEx()&MouseEvent.BUTTON2_DOWN_MASK)!=0) {
            //平行移動(Shift+で奥行も)
            if ((me.getModifiers()&MouseEvent.SHIFT_MASK)==0) {
                Tuple3d shift = a3canvas.canvasToVirtualCS(lastMouseX,lastMouseY,distance); 
                shift.sub(a3canvas.canvasToVirtualCS(me.getX(),me.getY(),distance));
                fixationPoint.add(shift);
            } else {
                Tuple3d shift = a3canvas.canvasToPhysicalCS(lastMouseX,lastMouseY,distance); 
                shift.sub(a3canvas.canvasToPhysicalCS(me.getX(),me.getY(),distance));
                shift.z = - shift.y;
                shift.y = 0.0;
                shift = a3canvas.physicalCSToVirtualCS(new Vector3d(shift));
                fixationPoint.add(shift);
            }
        } else if ((me.getModifiersEx()&MouseEvent.BUTTON3_DOWN_MASK)!=0) {
            //拡大縮小
            distance = distance*(100+shiftY)/100.0;
        }
        setCamera();
        lastMouseX = me.getX();
        lastMouseY = me.getY();
    }

    void setCamera() {
        Quat4d tmpQ = new Quat4d(0.0,0.0,1.0,0.0);
        Quat4d cameraQuatC = new Quat4d();
        cameraQuatC.conjugate(cameraQuat);
        tmpQ.mul(cameraQuat,tmpQ);
        tmpQ.mul(cameraQuatC);
        tmpQ.scale(distance);
        Vector3d tmpV = new Vector3d(fixationPoint);
        tmpV.set(tmpV.x+tmpQ.x,tmpV.y+tmpQ.y,tmpV.z+tmpQ.z);
        a3canvas.setCameraLocImmediately(tmpV);
        a3canvas.setCameraQuatImmediately(cameraQuat);
    }

    /**
     * マウスがリリースされた時のイベントをキャッチします。
     */
    public void mouseReleased(A3Event e) {
    }

    /**
     * マウスがクリックされた時のイベントをキャッチします。
     */
    public void mouseClicked(A3Event ae) {
    }

    /**
     * マウスがダブルクリックされた時のイベントをキャッチします。
     */
    public void mouseDoubleClicked(A3Event ae) {
        ;
    }

    /**
     * キーボードが押された時のイベントをキャッチします。
     */
    public void keyPressed(KeyEvent e) {
    }

    /**
     * キーボードが離された時のイベントをキャッチします。
     */
    public void keyReleased(KeyEvent e) {
    }

    /**
     * キーボードがタイプされた時のイベントをキャッチします。
     */
    public void keyTyped(KeyEvent e) {
    }
}
